import builtins
import copy

from paradigm import catalog
from . import examination
from .hints import Scope
from .leveling import to_flat_root


def to_children_scope(path: catalog.Path,
                      *,
                      scope: Scope) -> Scope:
    return {object_path: nodes[:]
            for object_path, nodes in scope.items()
            if object_path.is_child_of(path)
            and object_path != path}


def populate(module_path: catalog.Path,
             *,
             scope: Scope) -> None:
    root = to_flat_root(module_path)
    examination.conduct(root,
                        module_path=module_path,
                        scope=scope)


builtins_module_path = catalog.factory(builtins)
builtins_scope = {}
populate(builtins_module_path,
         scope=builtins_scope)


def factory(module_path: catalog.Path) -> Scope:
    result = copy.deepcopy(builtins_scope)
    if module_path != builtins_module_path:
        populate(module_path,
                 scope=result)
    return result



import ast
import importlib.machinery
import importlib.util
import os
import sys
from functools import singledispatch
from operator import methodcaller
from pathlib import Path
from types import (BuiltinFunctionType,
                   FunctionType,
                   ModuleType)
from typing import (Any,
                    Callable,
                    Iterable,
                    Optional,
                    Union)

from paradigm import catalog
from paradigm.hints import (MethodDescriptorType,
                            WrapperDescriptorType)
from . import unsupported


@singledispatch
def is_supported(object_: Any) -> bool:
    """
    Checks if object metadata extraction is supported.
    """
    raise TypeError('Unsupported object type: {type}.'
                    .format(type=type(object_)))


def find_stdlib_modules_names(directory_path: Path = Path(os.__file__).parent,
                              ) -> Iterable[str]:
    yield from sys.builtin_module_names

    def is_stdlib_module_path(path: Path) -> bool:
        base_name = path.stem
        # skips 'LICENSE', '__pycache__', 'site-packages', etc.
        return not (base_name.isupper()
                    or base_name.startswith('__')
                    or '-' in base_name)

    sources_paths = filter(is_stdlib_module_path, directory_path.iterdir())
    sources_relative_paths = map(methodcaller(Path.relative_to.__name__,
                                              directory_path),
                                 sources_paths)
    yield from map(str, map(methodcaller(Path.with_suffix.__name__, ''),
                            sources_relative_paths))


stdlib_modules_names = set(find_stdlib_modules_names())


@is_supported.register(ModuleType)
def is_module_supported(object_: ModuleType) -> bool:
    module_name = object_.__name__
    return (module_name in stdlib_modules_names
            and module_name not in unsupported.stdlib_modules_names
            and object_ not in unsupported.stdlib_modules
            or has_supported_python_source_file(object_))


@is_supported.register(Path)
def is_source_path_supported(source_path: Path) -> bool:
    try:
        source = source_path.read_text()
    except (FileNotFoundError, UnicodeDecodeError):
        return False
    try:
        ast.parse(source,
                  filename=str(source_path))
    except SyntaxError:
        return False
    return True


@is_supported.register(catalog.Path)
def is_module_path_supported(module_path: catalog.Path) -> bool:
    module_name = str(module_path)
    if module_name in stdlib_modules_names:
        if module_name in unsupported.stdlib_modules_names:
            return False
        module = importlib.import_module(module_name)
        return is_supported(module)
    spec = find_spec(module_path)
    if spec is None:
        return False
    source_path_string = spec.origin
    if source_path_string is None:
        return False
    return is_supported(Path(source_path_string))


def find_spec(module_path: catalog.Path
              ) -> Optional[importlib.machinery.ModuleSpec]:
    module_name = str(module_path)
    try:
        return importlib.util.find_spec(module_name)
    except (ImportError, ValueError):
        return None


@is_supported.register(BuiltinFunctionType)
def is_built_in_function_supported(object_: BuiltinFunctionType) -> bool:
    return (is_stdlib_callable_supported(object_)
            and is_not_private(object_)
            and has_module(object_)
            and object_ not in unsupported.built_in_functions)


@is_supported.register(type)
def is_class_supported(object_: type) -> bool:
    return (is_stdlib_callable_supported(object_)
            and is_not_private(object_)
            and object_ not in unsupported.classes
            or not is_stdlib_object(object_)
            and has_module(object_)
            and is_supported(catalog.factory(object_.__module__)))


@is_supported.register(FunctionType)
def is_function_supported(object_: FunctionType) -> bool:
    return (has_module(object_)
            and is_supported(catalog.factory(object_.__module__)))


@is_supported.register(MethodDescriptorType)
def is_method_descriptor_supported(object_: MethodDescriptorType) -> bool:
    return (is_stdlib_callable_supported(object_)
            and is_not_private(object_)
            and object_ not in unsupported.methods_descriptors)


@is_supported.register(WrapperDescriptorType)
def is_wrapper_descriptor_supported(object_: WrapperDescriptorType) -> bool:
    return (is_stdlib_callable_supported(object_)
            and object_ not in unsupported.wrappers_descriptors)


def has_supported_python_source_file(module: ModuleType) -> bool:
    try:
        file_path_string = module.__file__
    except AttributeError:
        return False
    return is_supported(Path(file_path_string))


def is_stdlib_object(object_: Any) -> bool:
    if not has_module(object_):
        return False
    top_module_name = catalog.factory(object_.__module__).parts[0]
    return top_module_name in stdlib_modules_names


def is_stdlib_callable_supported(callable_: Callable[..., Any]) -> bool:
    return callable_ not in unsupported.stdlib_modules_callables


def is_not_private(object_: Union[BuiltinFunctionType,
                                  MethodDescriptorType,
                                  type]) -> bool:
    return not object_.__name__.startswith('_')


def has_module(object_: Any) -> bool:
    return (hasattr(object_, '__module__')
            and object_.__module__)



from typing import (Any,
                    Dict,
                    List,
                    Union)

from typed_ast import ast3

from paradigm import catalog

Node = Union[ast3.AST, catalog.Path]
Scope = Dict[catalog.Path, List[Node]]
Namespace = Dict[str, Any]



import platform
import sys
from itertools import chain
from types import ModuleType
from typing import (Callable,
                    Iterable,
                    Union)

from .utils import to_contents

# importing will cause unwanted side effects such as raising error
stdlib_modules_names = {'antigravity', 'this'}

if sys.platform == 'win32':
    stdlib_modules_names.update({'crypt',
                                 'curses',
                                 'pty',
                                 'tty'})

if platform.python_implementation() == 'PyPy':
    stdlib_modules_names.update({'msilib',
                                 'symtable',
                                 'tracemalloc'})

stdlib_modules = set()

if platform.python_implementation() != 'PyPy':
    import _collections
    import _codecs_hk
    import _codecs_iso2022
    import _codecs_jp
    import _codecs_kr
    import _codecs_cn
    import _codecs_tw
    import _lsprof
    import _multibytecodec
    import _multiprocessing
    import _string
    import audioop
    import parser
    import xxsubtype

    # not supported by ``typeshed`` package
    stdlib_modules.update({_collections,
                           _codecs_hk,
                           _codecs_iso2022,
                           _codecs_jp,
                           _codecs_kr,
                           _codecs_cn,
                           _codecs_tw,
                           _lsprof,
                           _multibytecodec,
                           _multiprocessing,
                           _string,
                           audioop,
                           parser,
                           xxsubtype})

    if sys.version_info >= (3, 6):
        import _sha3

        stdlib_modules.add(_sha3)

    if ((3, 6) <= sys.version_info < (3, 6, 7)
            or (3, 7) <= sys.version_info < (3, 7, 1)):
        import _blake2

        stdlib_modules.add(_blake2)

    if sys.platform == 'win32':
        import _msi

        stdlib_modules.add(_msi)


def to_callables(object_: Union[ModuleType, type]) -> Iterable[Callable]:
    yield from filter(callable, to_contents(object_))


stdlib_modules_callables = list(chain.from_iterable(map(to_callables,
                                                        stdlib_modules)))

built_in_functions = set()

if platform.python_implementation() != 'PyPy':
    import _hashlib
    import _json
    import _thread
    import codecs
    import ctypes
    import socket

    # not supported by ``typeshed`` package
    built_in_functions.update({_hashlib.openssl_md5,
                               _hashlib.openssl_sha1,
                               _hashlib.openssl_sha224,
                               _hashlib.openssl_sha256,
                               _hashlib.openssl_sha384,
                               _hashlib.openssl_sha512,
                               _json.encode_basestring,
                               _thread.allocate,
                               _thread.exit_thread,
                               _thread.interrupt_main,
                               _thread.stack_size,
                               _thread.start_new_thread,
                               codecs.backslashreplace_errors,
                               codecs.ignore_errors,
                               codecs.namereplace_errors,
                               codecs.replace_errors,
                               codecs.strict_errors,
                               codecs.xmlcharrefreplace_errors,
                               ctypes._dlopen,
                               ctypes.pointer,
                               socket.dup,
                               sys.callstats,
                               sys.getallocatedblocks,
                               sys.get_coroutine_wrapper,
                               sys.set_coroutine_wrapper})
    if sys.version_info >= (3, 6):
        built_in_functions.update({sys.getfilesystemencodeerrors,
                                   sys.get_asyncgen_hooks,
                                   sys.set_asyncgen_hooks})

    if sys.version_info >= (3, 7):
        built_in_functions.update({socket.close,
                                   sys.breakpointhook})

    if sys.platform != 'win32':
        import _locale

        built_in_functions.update({_locale.bind_textdomain_codeset,
                                   _locale.bindtextdomain,
                                   _locale.dcgettext,
                                   _locale.dgettext,
                                   _locale.gettext,
                                   _locale.textdomain})

        if sys.version_info >= (3, 7):
            import time

            built_in_functions.add(time.pthread_getcpuclockid)

classes = set()

if platform.python_implementation() != 'PyPy':
    import _collections_abc
    import _io
    import _ssl
    import _thread
    import asyncio.events
    import ctypes
    import encodings
    import itertools
    import random
    import socket

    # not supported by ``typeshed`` package
    classes.update({_collections_abc.mappingproxy,
                    _io._BufferedIOBase,
                    _io._IOBase,
                    _io._RawIOBase,
                    _io._TextIOBase,
                    _ssl._SSLContext,
                    _thread.RLock,
                    _thread._local,
                    asyncio.events._RunningLoop,
                    ctypes._CFuncPtr,
                    itertools._grouper,
                    itertools._tee,
                    itertools._tee_dataobject,
                    encodings.CodecRegistryError,
                    random._MethodType})

    if sys.version_info < (3, 7):
        classes.add(_collections_abc.range_iterator)

    if sys.platform == 'win32':
        import msilib

        classes.update({msilib.UuidCreate,
                        msilib.FCICreate,
                        msilib.OpenDatabase,
                        msilib.CreateRecord})

        if sys.version_info < (3, 7):
            import os

            classes.update({os.uname_result,
                            os.statvfs_result})

methods_descriptors = set()

if platform.python_implementation() != 'PyPy':
    import _collections_abc
    import _io
    import _thread
    import collections

    # not supported by ``typeshed`` package
    methods_descriptors.update({_collections_abc.dict_items.isdisjoint,
                                _collections_abc.dict_keys.isdisjoint,
                                _collections_abc.generator.close,
                                _collections_abc.generator.send,
                                _collections_abc.generator.throw,
                                _collections_abc.coroutine.close,
                                _collections_abc.coroutine.send,
                                _collections_abc.coroutine.throw,
                                _io.BufferedRWPair.peek,
                                _thread.LockType.acquire_lock,
                                _thread.LockType.locked,
                                _thread.LockType.locked_lock,
                                _thread.LockType.release_lock,
                                collections.OrderedDict.clear,
                                collections.OrderedDict.pop,
                                collections.OrderedDict.update})

    if sys.version_info >= (3, 6):
        methods_descriptors.update({_collections_abc.async_generator.aclose,
                                    _collections_abc.async_generator.asend,
                                    _collections_abc.async_generator.athrow})
        if sys.platform == 'linux':
            import socket

            methods_descriptors.add(socket.socket.sendmsg_afalg)

    if sys.version_info >= (3, 7):
        import socket

        methods_descriptors.add(socket.socket.getblocking)
    else:
        import zipimport

        methods_descriptors.update({collections.OrderedDict.setdefault,
                                    zipimport.zipimporter.find_loader})

    if sys.platform == 'win32':
        import socket

        methods_descriptors.add(socket.socket.share)

wrappers_descriptors = set()

if platform.python_implementation() != 'PyPy':
    import _collections_abc

    # not supported by ``typeshed`` package
    wrappers_descriptors.update({
        _collections_abc.coroutine.__del__,
        _collections_abc.generator.__del__})

    if sys.version_info >= (3, 6):
        import _socket

        wrappers_descriptors.update({_collections_abc.async_generator.__del__,
                                     _socket.socket.__del__})



from typing import List

from paradigm import catalog
from .hints import (Node,
                    Scope)
from .logical import is_link


def search_nodes(path: catalog.Path,
                 *,
                 scope: Scope) -> List[Node]:
    nodes = scope[path]
    while is_link(nodes[-1]):
        nodes = scope[nodes[-1]]
    return nodes



import enum
from abc import (ABC,
                 abstractmethod)
from collections import defaultdict
from itertools import (chain,
                       takewhile)
from operator import (attrgetter,
                      methodcaller)
from typing import (Any,
                    Dict,
                    Iterable,
                    List,
                    Tuple)

from memoir import cached
from reprit.base import generate_repr

from .hints import Domain


class Parameter:
    class Kind(enum.IntEnum):
        POSITIONAL_ONLY = 0
        POSITIONAL_OR_KEYWORD = 1
        VARIADIC_POSITIONAL = 2
        KEYWORD_ONLY = 3
        VARIADIC_KEYWORD = 4

        def __repr__(self) -> str:
            return type(self).__qualname__ + '.' + self.name

        def __str__(self) -> str:
            return self.name.lower().replace('_', ' ')

    positionals_kinds = {Kind.POSITIONAL_ONLY, Kind.POSITIONAL_OR_KEYWORD}
    keywords_kinds = {Kind.POSITIONAL_OR_KEYWORD, Kind.KEYWORD_ONLY}
    kinds_prefixes = defaultdict(str,
                                 {Kind.VARIADIC_POSITIONAL: '*',
                                  Kind.VARIADIC_KEYWORD: '**'})

    def __init__(self,
                 *,
                 name: str,
                 kind: Kind,
                 has_default: bool) -> None:
        # performing validation inside of `__init__` instead of `__new__`,
        # because `pickle` does not support keyword only arguments in `__new__`
        if (kind not in (self.positionals_kinds | self.keywords_kinds)
                and has_default):
            raise ValueError('Variadic parameters '
                             'can\'t have default arguments.')
        self.name = name
        self.kind = kind
        self.has_default = has_default

    def __eq__(self, other: 'Parameter') -> bool:
        if not isinstance(other, Parameter):
            return NotImplemented
        return (self.name == other.name
                and self.kind == other.kind
                and self.has_default is other.has_default)

    def __hash__(self) -> int:
        return hash((self.name, self.kind, self.has_default))

    __repr__ = generate_repr(__init__)

    def __str__(self) -> str:
        return ''.join([self.kinds_prefixes[self.kind],
                        self.name,
                        '=...' if self.has_default else ''])


def to_parameters_by_kind(parameters: Iterable[Parameter]
                          ) -> Dict[Parameter.Kind, List[Parameter]]:
    result = defaultdict(list)
    for parameter in parameters:
        result[parameter.kind].append(parameter)
    return result


def to_parameters_by_name(parameters: Iterable[Parameter]
                          ) -> Dict[str, Parameter]:
    return {parameter.name: parameter
            for parameter in parameters}


def all_parameters_has_defaults(parameters: Iterable[Parameter]) -> bool:
    return all(map(attrgetter('has_default'), parameters))


class Base(ABC):
    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def __eq__(self, other: 'Base') -> bool:
        pass

    @abstractmethod
    def __hash__(self) -> int:
        pass

    @abstractmethod
    def all_set(self, *args: Domain, **kwargs: Domain) -> bool:
        pass

    @abstractmethod
    def expects(self, *args: Domain, **kwargs: Domain) -> bool:
        pass

    @abstractmethod
    def bind(self, *args: Domain, **kwargs: Domain) -> 'Base':
        pass


class Plain(Base):
    POSITIONAL_ONLY_SEPARATOR = '/'
    KEYWORD_ONLY_SEPARATOR = '*'

    def __new__(cls, *parameters: Parameter) -> 'Plain':
        try:
            prior, *rest = parameters
        except ValueError:
            pass
        else:
            visited_names = {prior.name}
            visited_kinds = {prior.kind}
            for parameter in rest:
                name = parameter.name

                if name in visited_names:
                    raise ValueError('Parameters should have unique names, '
                                     'but found duplicate '
                                     'for parameter "{name}".'
                                     .format(name=name))

                kind = parameter.kind

                if kind < prior.kind:
                    raise ValueError('Invalid parameters order: '
                                     'parameter "{prior_name}" '
                                     'with kind "{prior_kind!s}" '
                                     'precedes parameter "{parameter}" '
                                     'with kind "{kind!s}".'
                                     .format(prior_name=prior.name,
                                             prior_kind=prior.kind,
                                             kind=kind,
                                             parameter=name))

                if kind in Parameter.positionals_kinds:
                    if not parameter.has_default:
                        if prior.has_default:
                            raise ValueError('Invalid parameters order: '
                                             'parameter "{name}" '
                                             'without default argument '
                                             'follows '
                                             'parameter "{prior_name}" '
                                             'with default argument.'
                                             .format(name=name,
                                                     prior_name=prior.name))
                elif kind not in Parameter.keywords_kinds:
                    if kind in visited_kinds:
                        raise ValueError('Variadic parameters '
                                         'should have unique kinds, '
                                         'but found duplicate '
                                         'for kind "{kind!s}".'
                                         .format(kind=kind))

                prior = parameter
                visited_names.add(name)
                visited_kinds.add(kind)
        return super().__new__(cls)

    def __init__(self, *parameters: Parameter) -> None:
        self._parameters = parameters

    @property
    def parameters(self) -> Tuple[Parameter, ...]:
        return self._parameters

    def __eq__(self, other: Base) -> bool:
        if not isinstance(other, Base):
            return NotImplemented
        if not isinstance(other, Plain):
            return False
        return self._parameters == other._parameters

    def __hash__(self) -> int:
        return hash(self._parameters)

    __repr__ = generate_repr(__init__)

    def __str__(self) -> str:
        positionals_only = self.parameters_by_kind[
            Parameter.Kind.POSITIONAL_ONLY]
        parts = list(map(str, positionals_only))
        if positionals_only:
            parts.append(self.POSITIONAL_ONLY_SEPARATOR)
        parts.extend(map(str, self.parameters_by_kind[
            Parameter.Kind.POSITIONAL_OR_KEYWORD]))
        variadic_positionals = self.parameters_by_kind[
            Parameter.Kind.VARIADIC_POSITIONAL]
        parts.extend(map(str, variadic_positionals))
        keywords_only = self.parameters_by_kind[
            Parameter.Kind.KEYWORD_ONLY]
        if keywords_only and not variadic_positionals:
            parts.append(self.KEYWORD_ONLY_SEPARATOR)
        parts.extend(map(str, keywords_only))
        parts.extend(map(str, self.parameters_by_kind[
            Parameter.Kind.VARIADIC_KEYWORD]))
        return '(' + ', '.join(parts) + ')'

    @cached.property_
    def parameters_by_kind(self) -> Dict[Parameter.Kind, List[Parameter]]:
        return to_parameters_by_kind(self._parameters)

    def all_set(self, *args: Domain, **kwargs: Domain) -> bool:
        positionals = (self.parameters_by_kind[Parameter.Kind.POSITIONAL_ONLY]
                       + self.parameters_by_kind[
                           Parameter.Kind.POSITIONAL_OR_KEYWORD])
        unexpected_positional_arguments_found = (
                not self.parameters_by_kind[Parameter.Kind.VARIADIC_POSITIONAL]
                and len(args) > len(positionals))
        if unexpected_positional_arguments_found:
            return False
        rest_positionals = positionals[len(args):]
        rest_positionals_by_kind = to_parameters_by_kind(rest_positionals)
        rest_keywords = (
                rest_positionals_by_kind[Parameter.Kind.POSITIONAL_OR_KEYWORD]
                + self.parameters_by_kind[Parameter.Kind.KEYWORD_ONLY])
        rest_keywords_by_name = to_parameters_by_name(rest_keywords)
        unexpected_keyword_arguments_found = (
                not self.parameters_by_kind[Parameter.Kind.VARIADIC_KEYWORD]
                and kwargs.keys() - rest_keywords_by_name.keys())
        if unexpected_keyword_arguments_found:
            return False
        rest_keywords_by_name = {
            name: parameter
            for name, parameter in rest_keywords_by_name.items()
            if name not in kwargs}
        rest_positionals_only = (
            rest_positionals_by_kind[Parameter.Kind.POSITIONAL_ONLY])
        rest_keywords = rest_keywords_by_name.values()
        return (all_parameters_has_defaults(rest_positionals_only)
                and all_parameters_has_defaults(rest_keywords))

    def expects(self, *args: Domain, **kwargs: Domain) -> bool:
        positionals = (self.parameters_by_kind[Parameter.Kind.POSITIONAL_ONLY]
                       + self.parameters_by_kind[
                           Parameter.Kind.POSITIONAL_OR_KEYWORD])
        unexpected_positional_arguments_found = (
                not self.parameters_by_kind[Parameter.Kind.VARIADIC_POSITIONAL]
                and len(args) > len(positionals))
        if unexpected_positional_arguments_found:
            return False
        rest_positionals = positionals[len(args):]
        rest_positionals_by_kind = to_parameters_by_kind(rest_positionals)
        rest_keywords = (
                rest_positionals_by_kind[Parameter.Kind.POSITIONAL_OR_KEYWORD]
                + self.parameters_by_kind[Parameter.Kind.KEYWORD_ONLY])
        rest_keywords_by_name = to_parameters_by_name(rest_keywords)
        unexpected_keyword_arguments_found = (
                not self.parameters_by_kind[Parameter.Kind.VARIADIC_KEYWORD]
                and kwargs.keys() - rest_keywords_by_name.keys())
        if unexpected_keyword_arguments_found:
            return False
        return True

    def bind(self, *args: Domain, **kwargs: Domain) -> Base:
        return Plain(*_bind_keywords(_bind_positionals(
                self.parameters, args, kwargs,
                has_variadic=bool(self.parameters_by_kind[
                                      Parameter.Kind.VARIADIC_POSITIONAL])),
                kwargs,
                has_variadic=bool(self.parameters_by_kind[
                                      Parameter.Kind.VARIADIC_KEYWORD])))


def _bind_positionals(parameters: Tuple[Parameter, ...],
                      args: Tuple[Domain, ...],
                      kwargs: Dict[str, Domain],
                      *,
                      has_variadic: bool) -> Tuple[Parameter, ...]:
    def is_positional(parameter: Parameter) -> bool:
        return parameter.kind in Parameter.positionals_kinds

    positionals = tuple(takewhile(is_positional, parameters))
    if len(args) > len(positionals) and not has_variadic:
        value = 'argument' + 's' * (len(positionals) != 1)
        raise TypeError('Takes {parameters_count} positional {value}, '
                        'but {arguments_count} {verb} given.'
                        .format(parameters_count=len(positionals),
                                value=value,
                                arguments_count=len(args),
                                verb='was' if len(args) == 1 else 'were'))
    for positional in positionals[:len(args)]:
        if positional.name in kwargs:
            if positional.kind in Parameter.keywords_kinds:
                raise TypeError('Got multiple values '
                                'for parameter "{name}".'
                                .format(name=positional.name))
            else:
                raise TypeError('Parameter "{name}" is {kind!s}, '
                                'but was passed as a keyword.'
                                .format(name=positional.name,
                                        kind=positional.kind))
    return positionals[len(args):] + parameters[len(positionals):]


def _bind_keywords(parameters: Tuple[Parameter, ...],
                   kwargs: Dict[str, Any],
                   *,
                   has_variadic: bool) -> Tuple[Parameter, ...]:
    kwargs_names = set(kwargs)
    extra_kwargs_names = (kwargs_names -
                          {parameter.name
                           for parameter in parameters
                           if parameter.kind in Parameter.keywords_kinds})
    if extra_kwargs_names and not has_variadic:
        value = 'argument' + 's' * (len(extra_kwargs_names) != 1)
        raise TypeError('Got unexpected keyword {value} "{names}".'
                        .format(value=value,
                                names='", "'.join(extra_kwargs_names)))
    kwargs_names -= extra_kwargs_names
    if not kwargs_names:
        return parameters
    first_kwarg_index = next(index
                             for index, parameter in enumerate(parameters)
                             if parameter.name in kwargs_names)
    return (parameters[:first_kwarg_index]
            + tuple(Parameter(name=parameter.name,
                              kind=Parameter.Kind.KEYWORD_ONLY,
                              has_default=(parameter.has_default
                                           or parameter.name in kwargs_names))
                    if parameter.kind in Parameter.keywords_kinds
                    else parameter
                    for parameter in parameters[first_kwarg_index:]
                    if parameter.kind != Parameter.Kind.VARIADIC_POSITIONAL))


class Overloaded(Base):
    def __new__(cls, *signatures: Base) -> Base:
        if len(signatures) == 1:
            return signatures[0]
        return super().__new__(cls)

    def __init__(self, *signatures: Base) -> None:
        def flatten(signature: Base) -> Iterable[Base]:
            if isinstance(signature, type(self)):
                yield from signature.signatures
            else:
                yield signature

        self.signatures = tuple(chain.from_iterable(map(flatten, signatures)))

    def __eq__(self, other: Base) -> bool:
        if not isinstance(other, Base):
            return NotImplemented
        if not isinstance(other, Overloaded):
            return False
        return self.signatures == other.signatures

    def __hash__(self) -> int:
        return hash(self.signatures)

    __repr__ = generate_repr(__init__)

    def __str__(self) -> str:
        return ' or '.join(map(str, self.signatures))

    def all_set(self, *args: Domain, **kwargs: Domain) -> bool:
        return any(map(methodcaller(Base.all_set.__name__, *args, **kwargs),
                       self.signatures))

    def expects(self, *args: Domain, **kwargs: Domain) -> bool:
        return any(map(methodcaller(Base.expects.__name__, *args, **kwargs),
                       self.signatures))

    def bind(self, *args: Domain, **kwargs: Domain) -> Base:
        signatures = list(filter(methodcaller(Base.expects.__name__,
                                              *args, **kwargs),
                                 self.signatures))
        if not signatures:
            raise TypeError('No corresponding signature found.')
        signatures = map(methodcaller(Base.bind.__name__, *args, **kwargs),
                         signatures)
        return Overloaded(*signatures)



import typing
from functools import (partial,
                       singledispatch)
from typing import (NamedTuple,
                    Union,
                    overload)

from typed_ast import ast3

from paradigm import catalog
from .data_access import search_nodes
from .hints import (Node,
                    Scope)

TYPING_MODULE_PATH = catalog.factory(typing)
NAMED_TUPLE_CLASS_PATH = next(iter(catalog.paths_factory(NamedTuple)))
OVERLOAD_FUNCTION_PATH = next(iter(catalog.paths_factory(overload)))


@singledispatch
def evaluate_node(node: ast3.AST,
                  *,
                  scope: Scope,
                  module_path: catalog.Path) -> Node:
    raise TypeError('Unsupported node type: {type}.'
                    .format(type=type(node)))


@evaluate_node.register(ast3.Name)
def evaluate_name(node: ast3.Name,
                  *,
                  scope: Scope,
                  module_path: catalog.Path) -> Node:
    return catalog.factory(node.id)


@evaluate_node.register(ast3.expr)
def evaluate_ellipsis_or_num(node: Union[ast3.Ellipsis, ast3.Num],
                             *,
                             scope: Scope,
                             module_path: catalog.Path) -> Node:
    return catalog.factory(str(Ellipsis))


@evaluate_node.register(ast3.NameConstant)
def evaluate_name_constant(node: ast3.NameConstant,
                           *,
                           scope: Scope,
                           module_path: catalog.Path) -> Node:
    return catalog.factory(str(node.value))


@evaluate_node.register(ast3.Attribute)
def evaluate_attribute(node: ast3.Attribute,
                       *,
                       scope: Scope,
                       module_path: catalog.Path) -> Node:
    value_path = evaluate_node(node.value,
                               scope=scope,
                               module_path=module_path)
    return value_path.join(catalog.factory(node.attr))


@evaluate_node.register(ast3.Subscript)
def evaluate_subscript(node: ast3.Subscript,
                       *,
                       scope: Scope,
                       module_path: catalog.Path) -> Node:
    return evaluate_node(node.value,
                         scope=scope,
                         module_path=module_path)


@evaluate_node.register(ast3.Call)
def evaluate_call(node: ast3.Call,
                  *,
                  scope: Scope,
                  module_path: catalog.Path) -> Node:
    if not is_named_tuple_definition(node,
                                     scope=scope,
                                     module_path=module_path):
        return node
    class_name_node, fields_node = node.args

    def field_to_parameter(field_node: ast3.expr) -> ast3.arg:
        name_node, annotation_node = field_node.elts
        return ast3.arg(ast3.literal_eval(name_node), annotation_node)

    initializer_node = ast3.FunctionDef(
            '__init__',
            ast3.arguments([ast3.arg('self', None)]
                           + list(map(field_to_parameter,
                                      fields_node.elts)),
                           None, [], [], None, []),
            [ast3.Pass()], [], None)
    function_path = evaluate_node(node.func,
                                  scope=scope,
                                  module_path=module_path)
    class_def = ast3.ClassDef(ast3.literal_eval(class_name_node),
                              [ast3.Name(str(function_path), ast3.Load())],
                              [], [initializer_node], [])
    return ast3.fix_missing_locations(ast3.copy_location(class_def, node))


def is_named_tuple_definition(node: ast3.Call,
                              *,
                              scope: Scope,
                              module_path: catalog.Path) -> bool:
    evaluator = partial(evaluate_node,
                        scope=scope,
                        module_path=module_path)
    function_path = evaluator(node.func)
    return any_path_has_origin(function_path,
                               candidates_module_path=module_path,
                               candidates_scope=scope,
                               origin_module_path=TYPING_MODULE_PATH,
                               origin_object_path=NAMED_TUPLE_CLASS_PATH)


def is_overloaded_function(node: ast3.FunctionDef,
                           *,
                           scope: Scope,
                           module_path: catalog.Path) -> bool:
    evaluator = partial(evaluate_node,
                        scope=scope,
                        module_path=module_path)
    decorators_paths = map(evaluator, node.decorator_list)
    return any_path_has_origin(*decorators_paths,
                               candidates_module_path=module_path,
                               candidates_scope=scope,
                               origin_module_path=TYPING_MODULE_PATH,
                               origin_object_path=OVERLOAD_FUNCTION_PATH)


def any_path_has_origin(*candidates: catalog.Path,
                        candidates_module_path: catalog.Path,
                        candidates_scope: Scope,
                        origin_module_path: catalog.Path,
                        origin_object_path: catalog.Path) -> bool:
    searcher = partial(search_nodes,
                       scope=candidates_scope)
    if candidates_module_path == origin_module_path:
        return any(path == origin_object_path
                   for path in candidates)
    for candidate in candidates:
        try:
            candidate_nodes = searcher(candidate)
        except KeyError:
            if catalog.is_attribute(candidate):
                parent_nodes = searcher(candidate.parent)
                if isinstance(parent_nodes[-1], ast3.Import):
                    if any(to_actual_path(name_alias) == origin_module_path
                           for name_alias in parent_nodes.names):
                        return (catalog.factory(candidate.parts[-1])
                                == origin_object_path)
        else:
            origin_node = candidate_nodes[-1]
            if isinstance(origin_node, ast3.ImportFrom):
                if catalog.factory(origin_node.module) == origin_module_path:
                    try:
                        name_alias = next(
                                name_alias
                                for name_alias in reversed(origin_node.names)
                                if to_alias_path(name_alias) == candidate)
                    except StopIteration:
                        pass
                    else:
                        return (to_actual_path(name_alias)
                                == origin_object_path)
    return False


def to_actual_path(node: ast3.alias) -> catalog.Path:
    return catalog.factory(node.name)


def to_alias_path(node: ast3.alias) -> catalog.Path:
    return catalog.factory(to_alias_string(node))


def to_alias_string(node: ast3.alias) -> str:
    return node.asname or node.name



import builtins
import importlib
from collections import deque
from functools import (partial,
                       singledispatch)
from itertools import chain
from pathlib import Path
from typing import (Any,
                    Iterable,
                    List)

from typed_ast import ast3

from paradigm import (catalog,
                      sources)
from paradigm.hints import Predicate
from . import (construction,
               namespaces)
from .evaluation import (to_actual_path,
                         to_alias_path,
                         to_alias_string)
from .execution import execute
from .hints import Namespace


class ImportsRectifier(ast3.NodeTransformer):
    def __init__(self, module_path: catalog.Path) -> None:
        self.module_path = module_path

    def visit_Import(self, node: ast3.Import) -> Iterable[ast3.Import]:
        for name_alias in node.names:
            yield ast3.Import([name_alias])

    def visit_ImportFrom(self, node: ast3.ImportFrom
                         ) -> Iterable[ast3.ImportFrom]:
        parent_module_path = to_parent_module_path(
                node,
                parent_module_path=self.module_path)
        for name_alias in node.names:
            actual_path = to_actual_path(name_alias)
            if actual_path == catalog.WILDCARD_IMPORT:
                yield from to_flat_root(parent_module_path).body
            else:
                yield ast3.ImportFrom(str(parent_module_path), [name_alias], 0)

    def visit_FunctionDef(self, node: ast3.FunctionDef) -> ast3.FunctionDef:
        return node


def to_parent_module_path(object_: ast3.ImportFrom,
                          *,
                          parent_module_path: catalog.Path) -> catalog.Path:
    level = object_.level
    import_is_relative = level > 0
    if not import_is_relative:
        return catalog.factory(object_.module)
    depth = (len(parent_module_path.parts)
             + catalog.is_package(parent_module_path)
             - level) or None
    module_path_parts = filter(None,
                               chain(parent_module_path.parts[:depth],
                                     (object_.module,)))
    return catalog.Path(*module_path_parts)


class NamespaceUpdater(ast3.NodeVisitor):
    def __init__(self,
                 *,
                 namespace: Namespace,
                 module_path: catalog.Path,
                 parent_path: catalog.Path,
                 is_nested: bool) -> None:
        self.namespace = namespace
        self.module_path = module_path
        self.parent_path = parent_path
        self.is_nested = is_nested

    def visit_Import(self, node: ast3.Import) -> None:
        for name_alias in node.names:
            actual_path = to_actual_path(name_alias)
            parent_module_name = actual_path.parts[0]
            module = importlib.import_module(parent_module_name)
            self.namespace[parent_module_name] = module

    def visit_ImportFrom(self, node: ast3.ImportFrom) -> None:
        parent_module_path = to_parent_module_path(
                node,
                parent_module_path=self.module_path)
        for name_alias in node.names:
            alias_path = self.resolve_path(to_alias_path(name_alias))
            actual_path = to_actual_path(name_alias)
            if actual_path == catalog.WILDCARD_IMPORT:
                self.namespace.update(namespaces
                                      .from_module_path(parent_module_path))
                continue
            namespace = namespaces.from_module_path(parent_module_path)
            try:
                object_ = namespaces.search(namespace, actual_path)
            except KeyError:
                module_path = parent_module_path.join(actual_path)
                object_ = importlib.import_module(str(module_path))
            self.namespace[str(alias_path)] = object_

    def visit_ClassDef(self, node: ast3.ClassDef) -> None:
        path = self.resolve_path(catalog.factory(node.name))
        (NamespaceUpdater(namespace=self.namespace,
                          parent_path=path,
                          module_path=self.module_path,
                          is_nested=True)
         .generic_visit(node))

    def visit_FunctionDef(self, node: ast3.FunctionDef) -> None:
        return

    def resolve_path(self, path: catalog.Path) -> catalog.Path:
        if self.is_nested:
            return self.parent_path.join(path)
        return path


class IfsRectifier(ast3.NodeTransformer):
    def __init__(self,
                 *,
                 namespace: Namespace,
                 source_path: Path) -> None:
        self.namespace = namespace
        self.source_path = source_path

    def visit_FunctionDef(self, node: ast3.FunctionDef) -> ast3.FunctionDef:
        return node

    def visit_If(self, node: ast3.If) -> Iterable[ast3.AST]:
        if evaluate_expression(node.test,
                               source_path=self.source_path,
                               namespace=self.namespace):
            children = node.body
        else:
            children = node.orelse
        for child in children:
            self.generic_visit(child)
        yield from children


def evaluate_expression(node: ast3.expr,
                        *,
                        source_path: Path,
                        namespace: Namespace) -> Any:
    # to avoid name conflicts
    # we're using name that won't be present
    # because it'll lead to ``SyntaxError`` otherwise
    # and no AST will be generated
    temporary_name = '@tmp'
    assignment = expression_to_assignment(node,
                                          name=temporary_name)
    execute(assignment,
            source_path=source_path,
            namespace=namespace)
    return namespace.pop(temporary_name)


def expression_to_assignment(node: ast3.expr,
                             *,
                             name: str) -> ast3.Assign:
    name_node = ast3.copy_location(ast3.Name(name, ast3.Store()), node)
    result = ast3.Assign([name_node], node, None)
    return ast3.copy_location(result, node)


builtins_namespace = namespaces.from_module(builtins)


def to_flat_root(module_path: catalog.Path) -> ast3.Module:
    source_path = sources.factory(module_path)
    result = construction.from_source_path(source_path)
    flatten_root(result,
                 module_path=module_path,
                 source_path=source_path)
    return result


def flatten_root(module_root: ast3.Module,
                 *,
                 module_path: catalog.Path,
                 source_path: Path) -> None:
    flatten_imports(module_root,
                    module_path=module_path)
    flatten_ifs(module_root,
                module_path=module_path,
                source_path=source_path)


def flatten_imports(module_root: ast3.Module,
                    *,
                    module_path: catalog.Path) -> None:
    ImportsRectifier(module_path).visit(module_root)


def flatten_ifs(module_root: ast3.Module,
                *,
                module_path: catalog.Path,
                source_path: Path) -> None:
    namespace = dict(builtins_namespace)
    for node in left_search_within_children(module_root,
                                            ast3.If.__instancecheck__):
        dependencies_names = set()
        for child in left_search_within_children(node.test,
                                                 ast3.Name.__instancecheck__):
            if isinstance(child.ctx, ast3.Load):
                dependencies_names.add(child.id)
        update_namespace = NamespaceUpdater(namespace=namespace,
                                            module_path=module_path,
                                            parent_path=catalog.Path(),
                                            is_nested=False).visit
        while dependencies_names:
            dependency_name = dependencies_names.pop()
            dependency_node = next(right_search_within_children(
                    module_root,
                    partial(node_has_name,
                            name=dependency_name)))
            update_namespace(dependency_node)
    IfsRectifier(namespace=namespace,
                 source_path=source_path).visit(module_root)


def node_has_name(node: ast3.AST, name: str):
    return name in node_to_names(node)


@singledispatch
def node_to_names(node: ast3.AST) -> List[str]:
    return []


@node_to_names.register(ast3.ClassDef)
@node_to_names.register(ast3.FunctionDef)
def class_def_or_function_def_to_name(node: ast3.AST) -> List[str]:
    return [node.name]


@node_to_names.register(ast3.Import)
@node_to_names.register(ast3.ImportFrom)
def import_or_import_from_to_name(node: ast3.Import) -> List[str]:
    result = []
    for name_alias in node.names:
        result.append(to_alias_string(name_alias))
    return result


def left_search_within_children(node: ast3.AST,
                                condition: Predicate[ast3.AST]) -> Iterable:
    children = deque(ast3.iter_child_nodes(node))
    while children:
        child = children.popleft()
        if condition(child):
            yield child
        else:
            children.extend(ast3.iter_child_nodes(child))


def right_search_within_children(node, condition):
    children = deque(ast3.iter_child_nodes(node))
    while children:
        child = children.pop()
        if condition(child):
            yield child
        else:
            children.extendleft(ast3.iter_child_nodes(child))


builtins_root = to_flat_root(catalog.factory(builtins))



from functools import singledispatch
from itertools import (chain,
                       starmap)
from typing import (Any,
                    Dict,
                    Hashable,
                    List)

from typed_ast import ast3

from paradigm import catalog
from .hints import Node

is_link = catalog.Path.__instancecheck__
is_function_def = ast3.FunctionDef.__instancecheck__


@singledispatch
def are_similar(left_object: Any, right_object: Any) -> bool:
    return left_object == right_object


@are_similar.register(list)
def are_lists_similar(left_object: List[Any], right_object: List[Any]) -> bool:
    if type(left_object) is not type(right_object):
        return False
    if len(left_object) != len(right_object):
        return False
    return all(starmap(are_similar, zip(left_object, right_object)))


@are_similar.register(dict)
def are_dicts_similar(left_object: Dict[Hashable, Any],
                      right_object: Dict[Hashable, Any]) -> bool:
    if type(left_object) is not type(right_object):
        return False
    if left_object.keys() != right_object.keys():
        return False
    for key, left_value in left_object.items():
        if not are_similar(left_value, right_object[key]):
            return False
    return True


@are_similar.register(ast3.AST)
def are_nodes_similar(left_object: ast3.AST, right_object: Node) -> bool:
    if type(left_object) is not type(right_object):
        return False
    for property_name in chain(left_object._attributes, left_object._fields):
        try:
            left_property = getattr(left_object, property_name)
        except AttributeError:
            continue
        try:
            right_property = getattr(right_object, property_name)
        except AttributeError:
            return False
        if not are_similar(left_property, right_property):
            return False
    return True



from functools import partial
from typing import List

from typed_ast import ast3

from paradigm import (catalog,
                      sources)
from paradigm.hints import Map
from . import (examination,
               scoping)
from .data_access import search_nodes
from .evaluation import (evaluate_node,
                         to_actual_path,
                         to_alias_path)
from .hints import (Node,
                    Scope)
from .logical import is_link


def factory(*,
            module_path: catalog.Path,
            scope: Scope) -> Map[ast3.AST, None]:
    return Reducer(scope=scope,
                   module_path=module_path).visit


def complete_new_style_class_bases(bases: List[ast3.expr]) -> List[ast3.Expr]:
    return [*bases, ast3.Name('object', ast3.Load())]


class Reducer(ast3.NodeVisitor):
    def __init__(self,
                 *,
                 scope: Scope,
                 module_path: catalog.Path) -> None:
        self.scope = scope
        self.module_path = module_path

    def visit_Import(self, node: ast3.Import) -> None:
        for name_alias in node.names:
            alias_path = to_alias_path(name_alias)
            actual_path = to_actual_path(name_alias)
            for path, nodes in scoping.factory(actual_path).items():
                self.batch_register(alias_path.join(path), nodes)

    def visit_ImportFrom(self, node: ast3.ImportFrom) -> None:
        parent_module_path = catalog.factory(node.module)
        for name_alias in node.names:
            alias_path = to_alias_path(name_alias)
            actual_path = to_actual_path(name_alias)
            if actual_path == catalog.WILDCARD_IMPORT:
                for path, nodes in scoping.factory(parent_module_path).items():
                    self.batch_register_if_not_found(path, nodes)
                continue
            object_path = parent_module_path.join(actual_path)
            if is_module_path(object_path):
                for path, nodes in scoping.factory(object_path).items():
                    self.batch_register(alias_path.join(path), nodes)
            else:
                scope = scoping.factory(parent_module_path)
                target_nodes = scope[actual_path]
                while isinstance(target_nodes[-1],
                                 (ast3.Import, ast3.ImportFrom)):
                    visit_target = (Reducer(scope=scope,
                                            module_path=parent_module_path)
                                    .visit)
                    for target_node in target_nodes:
                        visit_target(target_node)
                    target_nodes = scope[actual_path]
                self.scope.update(scope)
                self.scope[alias_path] = target_nodes

    def visit_ClassDef(self, node: ast3.ClassDef) -> None:
        path = catalog.factory(node.name)
        bases = node.bases
        for base_index, base_node in enumerate(map(self.evaluator, bases)):
            if not is_link(base_node):
                base_scope = {}
                examination.conduct(base_node,
                                    scope=base_scope,
                                    module_path=self.module_path)
                for base_object_path, base_object_nodes in base_scope.items():
                    self.batch_register(base_object_path, base_object_nodes)
                base_class_path, = [
                    base_object_path
                    for base_object_path in base_scope
                    if not catalog.is_attribute(base_object_path)]
                bases[base_index] = ast3.Name(str(base_class_path),
                                              ast3.Load())
            else:
                try:
                    base_nodes = search_nodes(base_node,
                                              scope=self.scope)
                except KeyError:
                    while catalog.is_attribute(base_node):
                        base_node = base_node.parent
                        try:
                            base_nodes = search_nodes(base_node,
                                                      scope=self.scope)
                        except KeyError:
                            continue
                        else:
                            break
                    else:
                        raise
                self.visit(base_nodes[-1])
        bases = complete_new_style_class_bases(bases)
        for base_node in map(self.evaluator, bases):
            base_scope = scoping.to_children_scope(base_node,
                                                   scope=self.scope)
            for base_object_path, base_object_nodes in base_scope.items():
                self.batch_register_if_not_found(base_object_path
                                                 .with_parent(path),
                                                 base_object_nodes)
        children_scope = {}
        visit_child = Reducer(scope=children_scope,
                              module_path=path).visit
        for child in node.body:
            visit_child(child)
        for child_path, child_nodes in children_scope.items():
            self.batch_register(path.join(child_path), child_nodes)

    def batch_register(self, path, nodes: List[Node]) -> None:
        self.scope.setdefault(path, []).extend(nodes)

    def batch_register_if_not_found(self, path, nodes: List[Node]) -> None:
        self.scope.setdefault(path, nodes)

    @property
    def evaluator(self) -> Map[ast3.expr, catalog.Path]:
        return partial(evaluate_node,
                       scope=self.scope,
                       module_path=self.module_path)


def is_module_path(object_path: catalog.Path) -> bool:
    try:
        sources.factory(object_path)
    except KeyError:
        return False
    else:
        return True



from pathlib import Path

from typed_ast import ast3


def from_source_path(object_: Path) -> ast3.Module:
    return ast3.parse(object_.read_text())


def from_node(object_: ast3.AST) -> ast3.Module:
    return ast3.Module([object_], [])



import importlib
from functools import reduce
from types import ModuleType
from typing import Any

from paradigm import catalog
from .hints import Namespace


def from_module_path(object_: catalog.Path) -> Namespace:
    return from_module(importlib.import_module(str(object_)))


def from_module(object_: ModuleType) -> Namespace:
    return dict(vars(object_))


def contains(namespace: Namespace, path: catalog.Path) -> bool:
    try:
        search(namespace, path)
    except (KeyError, AttributeError):
        return False
    else:
        return True


def search(namespace: Namespace, path: catalog.Path) -> Any:
    return reduce(getattr, path.parts[1:], namespace[path.parts[0]])



import os
from itertools import (chain,
                       repeat,
                       starmap)
from operator import (itemgetter,
                      truediv)
from pathlib import Path
from typing import (Iterator,
                    List)

INIT_MODULE_NAME = '__init__'


def find_files(directory: Path) -> Iterator[Path]:
    def to_paths(root: str, files: List[str]) -> Iterator[Path]:
        yield from map(truediv, repeat(Path(root)), files)

    yield from chain.from_iterable(starmap(to_paths,
                                           map(itemgetter(0, 2),
                                               os.walk(str(directory)))))



import importlib
import inspect
import pathlib
import struct
import sys
import types
import weakref
from functools import singledispatch
from itertools import chain
from types import (BuiltinMethodType,
                   FunctionType,
                   ModuleType)
from typing import (Any,
                    Iterable,
                    Optional,
                    Union)

from memoir import cached

from .file_system import INIT_MODULE_NAME
from .hints import (MethodDescriptorType,
                    WrapperDescriptorType)


class Path:
    SEPARATOR = '.'

    def __init__(self, *parts: str) -> None:
        self.parts = parts

    def __str__(self) -> str:
        return self.SEPARATOR.join(self.parts)

    def __repr__(self) -> str:
        return (type(self).__qualname__
                + '(' + ', '.join(map(repr, self.parts)) + ')')

    def __eq__(self, other: 'Path') -> bool:
        if not isinstance(other, Path):
            return NotImplemented
        return self.parts == other.parts

    def __hash__(self) -> int:
        return hash(self.parts)

    def join(self, other: 'Path') -> 'Path':
        if not isinstance(other, Path):
            return NotImplemented
        return type(self)(*self.parts, *other.parts)

    @property
    def parent(self) -> 'Path':
        return type(self)(*self.parts[:-1])

    def with_parent(self, parent: 'Path') -> 'Path':
        return type(self)(*parent.parts, *self.parts[len(parent.parts):])

    def is_child_of(self, parent: 'Path') -> bool:
        return self.parts[:len(parent.parts)] == parent.parts


def is_attribute(path: Path) -> bool:
    return len(path.parts) > 1


WILDCARD_IMPORT = Path('*')


@singledispatch
def paths_factory(object_: Any) -> Iterable[Path]:
    yield factory(object_)


@paths_factory.register(BuiltinMethodType)
@paths_factory.register(FunctionType)
@paths_factory.register(MethodDescriptorType)
@paths_factory.register(WrapperDescriptorType)
@paths_factory.register(type)
def paths_from_class_or_function(object_: Union[BuiltinMethodType,
                                                FunctionType,
                                                MethodDescriptorType, type],
                                 *,
                                 max_depth: int = 2) -> Iterable[Path]:
    module = importlib.import_module(module_name_factory(object_))
    propertyspaces = iter(((Path(), module, 0),))
    while True:
        try:
            parent_path, propertyspace, depth = next(propertyspaces)
        except StopIteration:
            break

        def to_path(object_name: str) -> Path:
            return parent_path.join(factory(object_name))

        namespace = dict(vars(propertyspace))
        yield from (to_path(name)
                    for name, content in namespace.items()
                    if content is object_)
        next_depth = depth + 1
        if next_depth < max_depth:
            propertyspaces = chain(propertyspaces,
                                   [(to_path(name), content, next_depth)
                                    for name, content in namespace.items()
                                    if inspect.isclass(content)])
    yield from paths_factory(object_.__qualname__)


@singledispatch
def factory(object_: Any) -> Path:
    raise TypeError('Unsupported object type: {type}.'
                    .format(type=type(object_)))


@factory.register(ModuleType)
def from_module(object_: ModuleType) -> Path:
    return factory(object_.__name__)


@factory.register(pathlib.Path)
def from_relative_file_path(path: pathlib.Path) -> Path:
    if path.is_absolute():
        raise ValueError('Path should be relative.')
    *parts, module_file_name = path.parts

    def to_module_name(file_name: str) -> Optional[str]:
        if file_name == '.':
            return None
        result = inspect.getmodulename(file_name)
        if result == INIT_MODULE_NAME:
            return None
        return result

    module_name = to_module_name(module_file_name)
    if module_name is not None:
        parts = chain(parts, (module_name,))
    return Path(*parts)


names_replacements = {'Protocol': '_Protocol'}


@factory.register(str)
def from_string(string: str) -> Path:
    parts = string.split(Path.SEPARATOR)
    parts = map(names_replacements.get, parts, parts)
    return Path(*parts)


@singledispatch
def module_name_factory(object_: Any) -> str:
    raise TypeError('Unsupported object type: {type}.'
                    .format(type=type(object_)))


replacements = {'nt': 'os',
                'posix': 'os'}


@module_name_factory.register(str)
def module_name_from_string(object_: str) -> str:
    return replacements.get(object_, object_)


module_name_from_class_or_function_cache = {
    struct.Struct: struct.__name__,
    types.CodeType: types.__name__,
    types.FrameType: types.__name__,
    types.ModuleType: types.__name__,
    weakref.ref: weakref.__name__,
}

if sys.version_info >= (3, 7):
    import contextvars

    module_name_from_class_or_function_cache.update(
            {contextvars.Context: contextvars.__name__,
             contextvars.ContextVar: contextvars.__name__,
             contextvars.Token: contextvars.__name__,
             types.TracebackType: types.__name__})


@module_name_factory.register(BuiltinMethodType)
@module_name_factory.register(FunctionType)
@module_name_factory.register(type)
@cached.map_(types.MappingProxyType(module_name_from_class_or_function_cache))
def module_name_from_class_or_function(object_: Union[BuiltinMethodType,
                                                      FunctionType, type]
                                       ) -> str:
    result = object_.__module__
    if result is None:
        result = object_.__self__.__class__.__module__
    return module_name_factory(result)


@module_name_factory.register(MethodDescriptorType)
@module_name_factory.register(WrapperDescriptorType)
def module_name_from_method_descriptor(object_: MethodDescriptorType) -> str:
    return module_name_factory(object_.__objclass__)


def is_package(module_path: Path) -> bool:
    return hasattr(importlib.import_module(str(module_path)), '__path__')



from typing import (Callable,
                    TypeVar)

Domain = TypeVar('Domain')
Range = TypeVar('Range')
Map = Callable[[Domain], Range]
Operator = Map[Domain, Domain]
Predicate = Map[Domain, bool]

try:
    from types import MethodDescriptorType
except ImportError:
    MethodDescriptorType = type(list.append)
try:
    from types import WrapperDescriptorType
except ImportError:
    WrapperDescriptorType = type(list.__init__)



from functools import partial
from itertools import takewhile
from typing import List

from typed_ast import ast3

from paradigm import catalog
from . import (reduction,
               scoping)
from .data_access import search_nodes
from .evaluation import is_overloaded_function
from .logical import (are_dicts_similar,
                      are_lists_similar,
                      is_function_def,
                      is_link)


def to_functions_defs(object_path: catalog.Path,
                      module_path: catalog.Path) -> List[ast3.AST]:
    scope = scoping.factory(module_path)
    reduce_node = reduction.factory(module_path=module_path,
                                    scope=scope)
    while True:
        try:
            candidates = search_nodes(object_path,
                                      scope=scope)
        except KeyError:
            parent_path = object_path.parent
            parent_nodes = []
            while parent_path.parts:
                try:
                    parent_nodes = scope[parent_path]
                except KeyError:
                    parent_path = parent_path.parent
                else:
                    break
            if not parent_nodes:
                raise
            last_parent_node = parent_nodes[-1]
            if is_link(last_parent_node):
                object_path = object_path.with_parent(last_parent_node)
                continue
            children_scope_before = scoping.to_children_scope(parent_path,
                                                              scope=scope)
            reduce_node(last_parent_node)
            children_remain_intact = are_dicts_similar(
                    scoping.to_children_scope(parent_path,
                                              scope=scope),
                    children_scope_before)
            if children_remain_intact:
                return []
            continue
        if not is_function_def(candidates[-1]):
            nodes_before = candidates[:]
            reduce_node(candidates[-1])
            nodes_remain_intact = are_lists_similar(search_nodes(object_path,
                                                                 scope=scope),
                                                    nodes_before)
            if nodes_remain_intact:
                return []
            continue
        break
    last_candidate = candidates[-1]
    is_overload = partial(is_overloaded_function,
                          scope=scope,
                          module_path=module_path)
    if is_overload(last_candidate):
        return list(takewhile(is_overload, candidates))
    return [last_candidate]



from functools import singledispatch
from pathlib import Path

from typed_ast import ast3

from . import (construction,
               conversion)
from .hints import Namespace


@singledispatch
def execute(node: ast3.AST,
            *,
            source_path: Path,
            namespace: Namespace) -> None:
    raise TypeError('Unsupported node type: {type}.'
                    .format(type=type(node)))


@execute.register(ast3.stmt)
def execute_statement(node: ast3.stmt,
                      *,
                      source_path: Path,
                      namespace: Namespace) -> None:
    execute_tree(construction.from_node(node),
                 source_path=source_path,
                 namespace=namespace)


@execute.register(ast3.Module)
def execute_tree(node: ast3.Module,
                 *,
                 source_path: Path,
                 namespace: Namespace) -> None:
    code = compile(conversion.typed_to_plain(node), str(source_path), 'exec')
    exec(code, namespace)



from types import ModuleType
from typing import (Any,
                    List,
                    Union)


def to_contents(object_: Union[ModuleType, type]) -> List[Any]:
    return list(vars(object_).values())



from pathlib import Path

from setuptools import (find_packages,
                        setup)

import paradigm

project_base_url = 'https://github.com/lycantropos/paradigm/'

install_requires = [
    'memoir>=0.0.1',
    'reprit>=0.0.1',
    'typing_extensions>=3.6.5',
    'mypy>=0.700; platform_python_implementation != "PyPy"',
]
setup_requires = [
    'pytest-runner>=4.2',
]
tests_require = [
    'pytest>=3.8.1',
    'pytest-cov>=2.6.0',
    'hypothesis>=3.73.1',
]

setup(name='paradigm',
      packages=find_packages(exclude=('tests', 'tests.*',)),
      version=paradigm.__version__,
      classifiers=[
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: Implementation :: CPython',
          'Programming Language :: Python :: Implementation :: PyPy',
      ],
      description=paradigm.__doc__,
      long_description=Path('README.md').read_text(encoding='utf-8'),
      long_description_content_type='text/markdown',
      author='Azat Ibrakov',
      author_email='azatibrakov@gmail.com',
      url=project_base_url,
      download_url=project_base_url + 'archive/master.zip',
      python_requires='>=3.5.3',
      install_requires=install_requires,
      setup_requires=setup_requires,
      tests_require=tests_require)



import inspect
import sys
from collections import namedtuple
from functools import singledispatch
from itertools import chain
from pathlib import Path
from types import ModuleType
from typing import (Any,
                    Iterable,
                    Tuple)

from . import catalog
from .file_system import find_files
from .hints import Map

STUB_EXTENSION = '.pyi'


@singledispatch
def factory(object_: Any) -> Path:
    raise TypeError('Unsupported object type: {type}.'
                    .format(type=type(object_)))


@factory.register(ModuleType)
def from_module(object_: ModuleType) -> Path:
    try:
        return Path(object_.__path__[0])
    except AttributeError:
        return Path(inspect.getfile(object_))


@factory.register(catalog.Path)
def from_module_path(object_: catalog.Path) -> Path:
    try:
        return cache[object_]
    except KeyError as original_error:
        module_name_path = catalog.factory(object_.parts[-1].lstrip('_'))
        module_path = object_.parent.join(module_name_path)
        try:
            return cache[module_path]
        except KeyError:
            raise original_error


def generate_stubs_cache_items(root: Path
                               ) -> Iterable[Tuple[catalog.Path, Path]]:
    Version = namedtuple('Version', ['major', 'minor'])

    def is_supported_version_directory(directory: Path) -> bool:
        system_version = get_system_version()
        return any(version.major == system_version.major
                   and version.minor <= system_version.minor
                   for version in to_versions(directory))

    def to_versions(directory: Path,
                    *,
                    major_versions_separator: str = 'and'
                    ) -> Iterable[Version]:
        raw_versions = directory.name.split(major_versions_separator)

        def to_version(raw_version: str,
                       *,
                       version_parts_separator: str = '.') -> Version:
            raw_version_parts = raw_version.split(version_parts_separator)
            version_parts = list(map(int, raw_version_parts))
            try:
                major, minor = version_parts
            except ValueError:
                major, = version_parts
                minor = 0
            return Version(major, minor)

        yield from map(to_version, raw_versions)

    def get_system_version() -> Version:
        major, minor, *_ = sys.version_info
        return Version(major, minor)

    def module_full_name_factory(directory: Path) -> Map[Path, catalog.Path]:
        def to_module_path(stub: Path) -> catalog.Path:
            relative_stub_path = stub.relative_to(directory)
            return catalog.factory(relative_stub_path.with_suffix('.py'))

        return to_module_path

    def to_directory_items(directory: Path
                           ) -> Iterable[Tuple[catalog.Path, Path]]:
        to_module_full_name = module_full_name_factory(directory)
        stubs = filter(is_stub, find_files(directory))

        def to_directory_item(stub: Path) -> Tuple[catalog.Path, Path]:
            return to_module_full_name(stub), stub

        yield from map(to_directory_item, stubs)

    def to_directories_items(paths: Iterable[Path]
                             ) -> Iterable[Tuple[catalog.Path, Path]]:
        directories = filter(Path.is_dir, paths)
        directories = filter(is_supported_version_directory, directories)
        yield from chain.from_iterable(map(to_directory_items, directories))

    yield from to_directories_items(root.iterdir())


def is_stub(path: Path) -> bool:
    return path.suffixes == [STUB_EXTENSION]


try:
    import mypy
except ImportError:
    cache = {}
else:
    cache = dict(generate_stubs_cache_items(factory(mypy)
                                            / 'typeshed' / 'stdlib'))



import ast
from functools import partial
from itertools import chain
from typing import (Callable,
                    Iterable,
                    Optional,
                    Type)

from typed_ast import ast3

from paradigm.hints import Domain

TypedToPlainMethod = Callable[[ast3.NodeTransformer, ast3.AST],
                              Optional[ast.AST]]


def to_typed_to_plain_visitor(cls: Type[ast3.AST]) -> TypedToPlainMethod:
    try:
        plain_cls = getattr(ast, cls.__name__)
    except AttributeError:
        if issubclass(cls, ast3.stmt):
            def passer(_: ast3.NodeTransformer, node: cls) -> None:
                return ast.copy_location(ast.Pass(), node)

            return passer

        def deleter(_: ast3.NodeTransformer, __: cls) -> None:
            return None

        return deleter

    def visit(self: ast3.NodeTransformer, node: ast3.AST) -> ast.AST:
        node = self.generic_visit(node)
        result = plain_cls(*map(partial(getattr, node), plain_cls._fields))
        return ast3.copy_location(result, node)

    return visit


def to_subclasses(cls: Type[Domain],
                  *,
                  deep: bool = True) -> Iterable[Type[Domain]]:
    result = cls.__subclasses__()
    yield from result
    if not deep:
        return
    subclasses_factory = partial(to_subclasses,
                                 deep=deep)
    yield from chain.from_iterable(map(subclasses_factory, result))


class TypedToPlain(ast3.NodeTransformer):
    visitors = {'visit_' + cls.__name__: to_typed_to_plain_visitor(cls)
                for cls in set(to_subclasses(ast3.AST))}

    def __getattr__(self, name: str) -> TypedToPlainMethod:
        return partial(self.visitors[name], self)

    def generic_visit(self, node):
        for field, old_value in ast3.iter_fields(node):
            if isinstance(old_value, list):
                new_values = []
                for value in old_value:
                    if isinstance(value, ast3.AST):
                        value = self.visit(value)
                        if value is None:
                            continue
                        elif not isinstance(value, ast.AST):
                            new_values.extend(value)
                            continue
                    new_values.append(value)
                old_value[:] = new_values
            elif isinstance(old_value, ast3.AST):
                new_node = self.visit(old_value)
                if new_node is None:
                    delattr(node, field)
                else:
                    setattr(node, field, new_node)
        return node


typed_to_plain = TypedToPlain().visit



import inspect
import platform
from functools import (partial,
                       singledispatch,
                       wraps)
from itertools import (chain,
                       starmap,
                       zip_longest)
from operator import (attrgetter,
                      methodcaller)
from types import (BuiltinFunctionType,
                   BuiltinMethodType,
                   FunctionType,
                   MethodType)
from typing import (Any,
                    Callable,
                    Iterable,
                    Optional)
from weakref import WeakKeyDictionary

from memoir import cached

from .hints import (Map,
                    MethodDescriptorType,
                    WrapperDescriptorType)
from .models import (Base,
                     Overloaded,
                     Parameter,
                     Plain)


@singledispatch
def factory(object_: Callable[..., Any]) -> Base:
    raise TypeError('Unsupported object type: {type}.'
                    .format(type=type(object_)))


def from_callable(object_: Callable[..., Any]) -> Base:
    return from_raw_signature(inspect.signature(object_))


@factory.register(inspect.Signature)
def from_raw_signature(object_: inspect.Signature) -> Base:
    def normalize_parameter(raw_parameter: inspect.Parameter) -> Parameter:
        has_default = raw_parameter.default is not inspect._empty
        return Parameter(name=raw_parameter.name,
                         kind=Parameter.Kind(raw_parameter.kind),
                         has_default=has_default)

    parameters = map(normalize_parameter, object_.parameters.values())
    return Plain(*parameters)


from_class_cache = WeakKeyDictionary()
if platform.python_implementation() == 'PyPy':
    def from_class(object_: type) -> Base:
        try:
            return from_callable(object_)
        except ValueError:
            method = find_initializer_or_constructor(object_)
            if method is None:
                raise
            return slice_parameters(factory(method), slice(1, None))


    def find_initializer_or_constructor(class_: type) -> Callable:
        for cls, next_cls in zip(class_.__mro__, class_.__mro__[1:]):
            if cls.__init__ is not next_cls.__init__:
                return cls.__init__
            elif cls.__new__ is not next_cls.__new__:
                return cls.__new__
else:
    from typed_ast import ast3

    from . import (arboretum,
                   catalog)


    def with_typeshed(function: Map[Callable[..., Any], Base]
                      ) -> Map[Callable[..., Any], Base]:
        @wraps(function)
        def wrapped(object_: Callable[..., Any]) -> Base:
            try:
                return function(object_)
            except ValueError as error:
                object_paths = catalog.paths_factory(object_)
                module_path = catalog.factory(catalog
                                              .module_name_factory(object_))
                signature_factory = partial(to_signature,
                                            module_path=module_path)
                try:
                    return next(filter(None, map(signature_factory,
                                                 object_paths)))
                except StopIteration:
                    raise error

        return wrapped


    from_callable = with_typeshed(from_callable)

    from_class_cache[int] = Overloaded(
            Plain(Parameter(name='x',
                            kind=Parameter.Kind.POSITIONAL_ONLY,
                            has_default=True)),
            Plain(Parameter(name='x',
                            kind=Parameter.Kind.POSITIONAL_ONLY,
                            has_default=False),
                  Parameter(name='base',
                            kind=Parameter.Kind.POSITIONAL_OR_KEYWORD,
                            has_default=False)))


    def from_class(object_: type) -> Base:
        try:
            return from_callable(object_)
        except ValueError as error:
            method_paths = map(to_method_path, catalog.paths_factory(object_))
            module_path = catalog.factory(catalog.module_name_factory(object_))
            signature_factory = partial(to_signature,
                                        module_path=module_path)
            try:
                method_signature = next(filter(None, map(signature_factory,
                                                         method_paths)))
            except StopIteration:
                try:
                    base, = object_.__bases__
                except ValueError:
                    pass
                else:
                    if (object_.__init__ is base.__init__
                            and object_.__new__ is base.__new__):
                        result = from_class(base)
                        is_metaclass = base is type and object_ is not type
                        if is_metaclass:
                            if isinstance(result, Overloaded):
                                # metaclasses do not inherit
                                # single-argument constructor of `type`
                                result = max(result.signatures,
                                             key=to_max_parameters_count)
                        return result
                raise error
            else:
                return slice_parameters(method_signature, slice(1, None))


    to_method_path = methodcaller(catalog.Path.join.__name__,
                                  catalog.factory('__init__'))


    def to_signature(object_path: catalog.Path,
                     module_path: catalog.Path) -> Optional[Base]:
        try:
            nodes = arboretum.to_functions_defs(object_path, module_path)
        except KeyError:
            return None
        else:
            if not nodes:
                return None
            return Overloaded(*map(from_ast, map(attrgetter('args'), nodes)))


    def from_ast(signature_ast: ast3.arguments) -> Base:
        parameters = filter(
                None,
                chain(to_positional_parameters(signature_ast),
                      (to_variadic_positional_parameter(signature_ast),),
                      to_keyword_parameters(signature_ast),
                      (to_variadic_keyword_parameter(signature_ast),)))
        return Plain(*parameters)


    def to_positional_parameters(signature_ast: ast3.arguments
                                 ) -> Iterable[Parameter]:
        # double-reversing since parameters with default arguments go last
        parameters_with_defaults_ast = zip_longest(
                reversed(signature_ast.args),
                signature_ast.defaults)
        parameters_with_defaults_ast = reversed(
                list(parameters_with_defaults_ast))
        parameter_factory = partial(to_parameter,
                                    kind=Parameter.Kind.POSITIONAL_ONLY)
        yield from starmap(parameter_factory, parameters_with_defaults_ast)


    def to_keyword_parameters(signature_ast: ast3.arguments
                              ) -> Iterable[Parameter]:
        parameters_with_defaults_ast = zip(signature_ast.kwonlyargs,
                                           signature_ast.kw_defaults)
        parameter_factory = partial(to_parameter,
                                    kind=Parameter.Kind.KEYWORD_ONLY)
        yield from starmap(parameter_factory, parameters_with_defaults_ast)


    def to_variadic_positional_parameter(signature_ast: ast3.arguments
                                         ) -> Optional[Parameter]:
        parameter_ast = signature_ast.vararg
        if parameter_ast is None:
            return None
        return Parameter(name=parameter_ast.arg,
                         kind=Parameter.Kind.VARIADIC_POSITIONAL,
                         has_default=False)


    def to_variadic_keyword_parameter(signature_ast: ast3.arguments
                                      ) -> Optional[Parameter]:
        parameter_ast = signature_ast.kwarg
        if parameter_ast is None:
            return None
        return Parameter(name=parameter_ast.arg,
                         kind=Parameter.Kind.VARIADIC_KEYWORD,
                         has_default=False)


    def to_parameter(parameter_ast: ast3.arg,
                     default_ast: Optional[ast3.expr],
                     *,
                     kind: Parameter.Kind) -> Parameter:
        return Parameter(name=parameter_ast.arg,
                         kind=kind,
                         has_default=default_ast is not None)

from_callable = [factory.register(cls, from_callable)
                 for cls in (BuiltinFunctionType,
                             BuiltinMethodType,
                             FunctionType,
                             MethodType,
                             MethodDescriptorType,
                             WrapperDescriptorType)][-1]
from_class = factory.register(type)(cached.map_(from_class_cache)(from_class))


@factory.register(partial)
def from_partial(object_: partial) -> Base:
    return factory(object_.func).bind(*object_.args, **object_.keywords)


@singledispatch
def slice_parameters(signature: Base,
                     slice_: slice) -> Base:
    raise TypeError('Unsupported signature type: {type}.'
                    .format(type=type(signature)))


@slice_parameters.register(Plain)
def slice_plain_parameters(signature: Plain,
                           slice_: slice) -> Base:
    return Plain(*signature.parameters[slice_])


@slice_parameters.register(Overloaded)
def slice_overloaded_parameters(signature: Overloaded,
                                slice_: slice) -> Base:
    return Overloaded(*map(partial(slice_parameters,
                                   slice_=slice_),
                           signature.signatures))


@singledispatch
def to_max_parameters_count(signature: Base) -> int:
    raise TypeError('Unsupported signature type: {type}.'
                    .format(type=type(signature)))


@to_max_parameters_count.register(Plain)
def to_max_plain_parameters_count(signature: Plain) -> int:
    return len(signature.parameters)


@to_max_parameters_count.register(Overloaded)
def to_max_overloaded_parameters_count(signature: Overloaded) -> int:
    return max(map(to_max_parameters_count, signature.signatures))



from functools import partial
from typing import List

from typed_ast import ast3

from paradigm import catalog
from paradigm.hints import Map
from .evaluation import (evaluate_node,
                         to_alias_path)
from .hints import (Node,
                    Scope)
from .logical import is_link


class Registry(ast3.NodeVisitor):
    def __init__(self,
                 *,
                 scope: Scope,
                 module_path: catalog.Path) -> None:
        self.scope = scope
        self.module_path = module_path

    def visit_Import(self, node: ast3.Import):
        for name_alias in node.names:
            self.register(to_alias_path(name_alias), node)

    def visit_ImportFrom(self, node: ast3.ImportFrom):
        for name_alias in node.names:
            self.register(to_alias_path(name_alias), node)

    def visit_FunctionDef(self, node: ast3.FunctionDef):
        self.generic_visit(node)
        path = catalog.factory(node.name)
        self.register(path, node)

    def visit_ClassDef(self, node: ast3.ClassDef):
        path = catalog.factory(node.name)
        self.register(path, node)
        children_scope = {}
        visit_child = Registry(scope=children_scope,
                               module_path=self.module_path).visit
        for child in node.body:
            visit_child(child)
        for child_path, child_nodes in children_scope.items():
            self.batch_register(path.join(child_path), child_nodes)

    def visit_Assign(self, node: ast3.Assign):
        value_node = self.evaluator(node.value)
        for path in map(self.evaluator, node.targets):
            self.register(path, value_node)
        if not is_link(value_node):
            self.visit(value_node)

    def visit_AnnAssign(self, node: ast3.AnnAssign):
        self.register(self.evaluator(node.target), node)

    def batch_register(self, path, nodes: List[Node]) -> None:
        self.scope.setdefault(path, []).extend(nodes)

    def register(self, path, node: Node) -> None:
        self.scope.setdefault(path, []).append(node)

    @property
    def evaluator(self) -> Map[ast3.expr, catalog.Path]:
        return partial(evaluate_node,
                       scope=self.scope,
                       module_path=self.module_path)


def conduct(node: ast3.AST,
            *,
            module_path: catalog.Path,
            scope: Scope) -> None:
    Registry(scope=scope,
             module_path=module_path).visit(node)
